
class StoreLocalStorage {
  constructor(key) {
    this.key = key;
    this.loaded = false;
    this.loading = false;
    this.persisting = false;
  }

  persist(data) {
    return new Promise((resolve, reject) => {
      if (this.persistTimeout) {
        clearTimeout(this.persistTimeout);
      }
      this.persistTimeout = setTimeout(() => {
        this.persistTimeout = null;
        if (this.persisting || this.loading) {
          this.persist();
          return;
        }
        this.persisting = true;
        var key = typeof(this.key) === 'function' ? this.key() : this.key;
        StoreLocalStorage.db.setItem(key, data).then(() => {
          this.persisting = false;
          resolve();
        }).catch(reject);
      }, 1000);
    });
  }

  load() {
    if (this.loading) return this.ready;
    this.loading = true;
    this.ready = new Promise((resolve, reject) => {
      var key = typeof(this.key) === 'function' ? this.key() : this.key;
      StoreLocalStorage.db.getItem(key).then((objects) => {
        this.loaded = true;
        this.loading = false;
        resolve(objects);
      }).catch(reject);
    });
    return this.ready;
  }
}

StoreLocalStorage.db = localforage.createInstance({name: "StoreV1"});


function StoreDataLocalStorageWrapper(data, key, autoload, useReplaceAfterLoad) {
  this.data = data;
  this.localStorage = new StoreLocalStorage(key);
  this.useReplaceAfterLoad = useReplaceAfterLoad === undefined ? true : useReplaceAfterLoad;
  if (autoload === undefined ? true : autoload) {
    this.load();
  }
}

StoreDataLocalStorageWrapper.prototype.load = function() {
  return this.localStorage.load().then(objects => {
    if (objects) {
      try {
        if (this.useReplaceAfterLoad) {
          this.data.replace(objects);
        } else {
          objects.forEach(this.data.push.bind(this.data));
        }
      } catch(e) {
        console.log(e);
      }
    }
  })
};

StoreDataLocalStorageWrapper.prototype.persist = function() {
  return this.localStorage.persist(this.data.all());
};

['exists', 'count', 'all', 'get', 'ids', '_index', 'getIndexes', 'getIndexDimensions'].forEach(function(name) {
  StoreDataLocalStorageWrapper.prototype[name] = function() {
    if (!this.loaded && this.autoload) {
      this.load();
    }
    return this.data[name].apply(this.data, Array.prototype.slice.call(arguments, 0));
  };
});

['push', 'update', '_index', 'remove', 'replace'].forEach(function(name) {
  StoreDataLocalStorageWrapper.prototype[name] = function() {
    var r = this.data[name].apply(this.data, Array.prototype.slice.call(arguments, 0));
    this.persist();
    return r;
  };
});


StoreLocalStorageMixin = {
  enableLocalStorage: true,
  localStorageKey: null,
  localStorageScopedKey: false,
  localStorageAutoload: true,
  localStorageReplace: true,
  _createKeyFromScope: function() {
    var scopeStr = JSON.stringify(this.scopeFilters);
    var hash = 0, i, chr;
    if (scopeStr.length === 0) return hash;
    for (i = 0; i < scopeStr.length; i++) {
      chr   = scopeStr.charCodeAt(i);
      hash  = ((hash << 5) - hash) + chr;
      hash |= 0; // Convert to 32bit integer
    }
    return hash;
  },
  _getLocalStorageKey: function() {
    var key = this.localStorageKey;
    if (!key && this.typeName) {
      key = this.typeName + '-store';
    } else if (!key) {
      console.warn('Missing localStorageKey for store, localStorage is disabled');
      return null;
    }
    if (this.localStorageScopedKey && this.scopeFilters) {
      key += '-' + this._createKeyFromScope();
    }
    return key;
  },
  _wrapWithLocalStorage: function(data) {
    var key = this._getLocalStorageKey();
    if (!this.enableLocalStorage || !key) {
      return data;
    }
    return new StoreDataLocalStorageWrapper(data, key, this.localStorageAutoload, this.localStorageReplace);
  },
  _createData: function() {
    return this._wrapWithLocalStorage(Store.prototype._createData.call(this));
  },
  resetWithoutLocalStorage: function() {
    this.enableLocalStorage = false;
    this._clear(true);
  },
  createSnapshot: function() {
    var key = this._getLocalStorageKey();
    if (!key) {
      return Promise.resolve();
    }
    return new StoreLocalStorage(key).persist(this.query().toArray());
  },
  restoreSnapshot: function() {
    var key = this._getLocalStorageKey();
    if (!key) {
      return Promise.resolve();
    }
    return new StoreLocalStorage(key).load().then(objects => {
      if (objects) {
        const data = this.root().data;
        objects.forEach(obj => {
          if (!data.exists(obj.id)) {
            data.push(obj);
          }
        });
      }
    });
  },
  tryGetFromStore: function(id) {
    try {
      return Promise.resolve(this.get(id));
    } catch (e) {
      if (this.enableLocalStorage || !(e instanceof StoreNotFoundError)) {
        return Promise.reject(e);
      }
      return this.getFromStore(id);
    }
  },
  getFromStore: function(id) {
    var key = this._getLocalStorageKey();
    if (!key) {
      return Promise.reject();
    }
    return new StoreLocalStorage(key).load().then(objects => {
      if (objects) {
        let obj = objects.find(obj => obj.id === id);
        if (obj) {
          return obj;
        }
      }
      return Promise.reject();
    });
  }
};

StoreLoaderLocalStorageMixin = {
  _getOrLoad(id, action, params, callback, errorCallback) {
    this.store.tryGetFromStore(id).then(obj => {
      if (typeof(params) === 'function') {
        params(obj);
      } else if (callback) {
        callback(obj);
      }
    }).catch(() => {
      this._load(action, params, callback, errorCallback);
    });
  }
};
